@@ -212,6 +212,7 @@ compile_dyninst (systemtap_session& s)
     cmd.push_back("-D" + s.c_macros[i]);
   cmd.push_back(s.translated_source);
   cmd.push_back("-pthread");
+  cmd.push_back("-lrt");
   cmd.push_back("-fPIC");
   cmd.push_back("-shared");
   cmd.push_back("-o");
@@ -364,7 +365,11 @@ compile_pass (systemtap_session& s)
 			   "STAPCONF_UPROBE_REGISTER_EXPORTED");
   output_either_exportconf(s, o, "uprobe_unregister", "unregister_uprobe",
 			   "STAPCONF_UPROBE_UNREGISTER_EXPORTED");
+  output_exportconf(s, o, "uretprobe_register", "STAPCONF_URETPROBE_REGISTER_EXPORTED");
+  output_exportconf(s, o, "uretprobe_unregister", "STAPCONF_URETPROBE_UNREGISTER_EXPORTED");
   output_autoconf(s, o, "autoconf-old-inode-uprobes.c", "STAPCONF_OLD_INODE_UPROBES", NULL);
+  output_autoconf(s, o, "autoconf-inode-uprobes-noaddr.c", "STAPCONF_INODE_UPROBES_NOADDR", NULL);
+  output_autoconf(s, o, "autoconf-inode-uretprobes.c", "STAPCONF_INODE_URETPROBES", NULL);
 
   // used by tapsets.cxx inode uprobe generated code
   output_exportconf(s, o, "uprobe_get_swbp_addr", "STAPCONF_UPROBE_GET_SWBP_ADDR_EXPORTED");
@@ -374,6 +379,9 @@ compile_pass (systemtap_session& s)
 
   // used by runtime/stp_utrace.c
   output_exportconf(s, o, "task_work_add", "STAPCONF_TASK_WORK_ADD_EXPORTED");
+  output_exportconf(s, o, "signal_wake_up_state", "STAPCONF_SIGNAL_WAKE_UP_STATE_EXPORTED");
+  output_exportconf(s, o, "signal_wake_up", "STAPCONF_SIGNAL_WAKE_UP_EXPORTED");
+  output_exportconf(s, o, "__lock_task_sighand", "STAPCONF___LOCK_TASK_SIGHAND_EXPORTED");
 
   output_autoconf(s, o, "autoconf-pagefault_disable.c", "STAPCONF_PAGEFAULT_DISABLE", NULL);
   output_exportconf(s, o, "kallsyms_lookup_name", "STAPCONF_KALLSYMS");
@@ -599,7 +607,7 @@ uprobes_pass (systemtap_session& s)
 
   if (s.kernel_config["CONFIG_UTRACE"] != string("y"))
     {
-      clog << _("user-space facilities not available without kernel CONFIG_UTRACE or CONFIG_TRACEPOINTS/CONFIG_ARCH_SUPPORTS_UPROBES/CONFIG_UPROBES") << endl;
+      clog << _("user-space process-tracking facilities not available [man error::process-tracking]") << endl;
       s.set_try_server ();
       return 1;
     }
@@ -628,9 +636,9 @@ make_dyninst_run_command (systemtap_session& s, const string& remotedir,
 {
   vector<string> cmd;
   cmd.push_back(getenv("SYSTEMTAP_STAPDYN") ?: BINDIR "/stapdyn");
-  if (s.verbose > 1)
-    cmd.push_back("-v");
-  if (s.verbose > 2)
+
+  // use slightly less verbosity
+  for (unsigned i=1; i<s.verbose; i++)
     cmd.push_back("-v");
   if (s.suppress_warnings)
     cmd.push_back("-w");
@@ -641,6 +649,12 @@ make_dyninst_run_command (systemtap_session& s, const string& remotedir,
       cmd.push_back(s.cmd);
     }
 
+  if (s.target_pid)
+    {
+      cmd.push_back("-x");
+      cmd.push_back(lex_cast(s.target_pid));
+    }
+
   cmd.push_back((remotedir.empty() ? s.tmpdir : remotedir)
 		+ "/" + s.module_filename());
 
@@ -657,9 +671,9 @@ make_run_command (systemtap_session& s, const string& remotedir,
   // for now, just spawn staprun
   vector<string> staprun_cmd;
   staprun_cmd.push_back(getenv("SYSTEMTAP_STAPRUN") ?: BINDIR "/staprun");
-  if (s.verbose>1)
-    staprun_cmd.push_back("-v");
-  if (s.verbose>2)
+
+  // use slightly less verbosity
+  for (unsigned i=1; i<s.verbose; i++)
     staprun_cmd.push_back("-v");
   if (s.suppress_warnings)
     staprun_cmd.push_back("-w");
@@ -769,6 +783,11 @@ make_tracequeries(systemtap_session& s, const map<string,string>& contents)
       osrc << src;
       osrc.close();
 
+      if (s.verbose > 2)
+        clog << _F("Processing tracepoint header %s with query %s", 
+                   it->first.c_str(), srcname.c_str())
+             << endl;
+
       // arrange to build it
       omf << "obj-m += " + sbasename + ".o" << endl; // NB: without <dir> prefix
       objs[it->first] = dir + "/" + sbasename + ".o";
