@@ -22,6 +22,7 @@
 #include <linux/ptrace.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/time.h>
 #include <fcntl.h>
 #include <stddef.h>
 #include <unistd.h>
@@ -42,8 +43,23 @@ typedef uint32_t u32;
 typedef uint64_t u64;
 
 #include "linux_types.h"
+#include "offset_list.h"
 
 
+#ifndef NSEC_PER_SEC
+#define NSEC_PER_SEC 1000000000L
+#endif
+
+#define _stp_timespec_sub(a, b, result)					      \
+  do {									      \
+    (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;			      \
+    (result)->tv_nsec = (a)->tv_nsec - (b)->tv_nsec;			      \
+    if ((result)->tv_nsec < 0) {					      \
+      --(result)->tv_sec;						      \
+      (result)->tv_nsec += NSEC_PER_SEC;				      \
+    }									      \
+  } while (0)
+
 #define simple_strtol strtol
 
 
@@ -70,34 +86,17 @@ static inline int pseudo_atomic_cmpxchg(atomic_t *v, int oldval, int newval)
 	return __sync_val_compare_and_swap(&(v->counter), oldval, newval);
 }
 
-
-static pthread_mutex_t stapdyn_big_dumb_lock = PTHREAD_MUTEX_INITIALIZER;
-
-static inline void _stp_runtime_entryfn_prologue(void)
-{
-	pthread_mutex_lock(&stapdyn_big_dumb_lock);
-}
-
-static inline void _stp_runtime_entryfn_epilogue(void)
-{
-	pthread_mutex_unlock(&stapdyn_big_dumb_lock);
-}
-
-
 #include "linux_defs.h"
 
 #define MODULE_DESCRIPTION(str)
 #define MODULE_LICENSE(str)
 #define MODULE_INFO(tag,info)
 
-/* XXX for now, act like uniprocessor... */
-#define NR_CPUS 1
-#define num_online_cpus() 1
-#define smp_processor_id() 0
-#define get_cpu() 0
-#define put_cpu() 0
-#define for_each_possible_cpu(cpu) for ((cpu) = 0; (cpu) < NR_CPUS; ++(cpu))
-#define stp_for_each_cpu(cpu) for_each_possible_cpu((cpu))
+/* Semi-forward declaration from runtime_context.h, needed by stat.c. */
+static int _stp_runtime_num_contexts;
+
+#define for_each_possible_cpu(cpu) for ((cpu) = 0; (cpu) < _stp_runtime_num_contexts; (cpu)++)
+
 #define yield() sched_yield()
 
 #define access_ok(type, addr, size) 1
@@ -118,22 +117,82 @@ static inline void _stp_runtime_entryfn_epilogue(void)
 
 #include "io.c"
 #include "alloc.c"
+#include "shm.c"
 #include "print.h"
 #include "stp_string.c"
 #include "arith.c"
 #include "copy.c"
-#include "regs.c"
-#include "regs-ia64.c"
+#include "../regs.c"
 #include "task_finder.c"
 #include "sym.c"
 #include "perf.c"
 #include "addr-map.c"
+#include "stat.c"
 #include "unwind.c"
 
 static int systemtap_module_init(void);
 static void systemtap_module_exit(void);
 
-static unsigned long stap_hash_seed; /* Init during module startup */
+static inline unsigned long _stap_hash_seed(); /* see common_session_state.h */
+#define stap_hash_seed _stap_hash_seed()
+
+typedef pthread_rwlock_t rwlock_t; /* for globals */
+
+static int stp_pthread_mutex_init_shared(pthread_mutex_t *mutex)
+{
+	int rc;
+	pthread_mutexattr_t attr;
+
+	rc = pthread_mutexattr_init(&attr);
+	if (rc != 0) {
+	    _stp_error("pthread_mutexattr_init failed");
+	    return rc;
+	}
+
+	rc = pthread_mutexattr_setpshared(&attr, PTHREAD_PROCESS_SHARED);
+	if (rc != 0) {
+	    _stp_error("pthread_mutexattr_setpshared failed");
+	    goto err_attr;
+	}
+
+	rc = pthread_mutex_init(mutex, &attr);
+	if (rc != 0) {
+	    _stp_error("pthread_mutex_init failed");
+	    goto err_attr;
+	}
+
+err_attr:
+	(void)pthread_mutexattr_destroy(&attr);
+	return rc;
+}
+
+static int stp_pthread_rwlock_init_shared(pthread_rwlock_t *rwlock)
+{
+	int rc;
+	pthread_rwlockattr_t attr;
+
+	rc = pthread_rwlockattr_init(&attr);
+	if (rc != 0) {
+	    _stp_error("pthread_rwlockattr_init failed");
+	    return rc;
+	}
+
+	rc = pthread_rwlockattr_setpshared(&attr, PTHREAD_PROCESS_SHARED);
+	if (rc != 0) {
+	    _stp_error("pthread_rwlockattr_setpshared failed");
+	    goto err_attr;
+	}
+
+	rc = pthread_rwlock_init(rwlock, &attr);
+	if (rc != 0) {
+	    _stp_error("pthread_rwlock_init failed");
+	    goto err_attr;
+	}
+
+err_attr:
+	(void)pthread_rwlockattr_destroy(&attr);
+	return rc;
+}
 
 
 /*
@@ -161,17 +220,24 @@ static unsigned long stap_hash_seed; /* Init during module startup */
  * XXX Functions like _exit() which bypass destructors are a problem...
  */
 
+static int _stp_runtime_contexts_init(void);
+
 static pid_t stp_dyninst_master = 0;
 
+static int stp_dyninst_ctor_rc = 0;
+
 __attribute__((constructor))
 static void stp_dyninst_ctor(void)
 {
-    stap_hash_seed = _stp_random_u ((unsigned long)-1);
+    int rc = 0;
 
     _stp_mem_fd = open("/proc/self/mem", O_RDWR /*| O_LARGEFILE*/);
     if (_stp_mem_fd != -1) {
         fcntl(_stp_mem_fd, F_SETFD, FD_CLOEXEC);
     }
+    else {
+        rc = -errno;
+    }
 
     /* XXX We don't really want to be using the target's stdio. (PR14491)
      * But while we are, clone our own FILE handles so we're not affected by
@@ -179,18 +245,49 @@ static void stp_dyninst_ctor(void)
      */
     _stp_out = _stp_clone_file(stdout);
     _stp_err = _stp_clone_file(stderr);
+
+    if (rc == 0)
+        rc = _stp_runtime_contexts_init();
+
+    if (rc == 0)
+        rc = _stp_print_init();
+
+    stp_dyninst_ctor_rc = rc;
+}
+
+const char* stp_dyninst_shm_init(void)
+{
+    return _stp_shm_init();
+}
+
+int stp_dyninst_shm_connect(const char* name)
+{
+    /* We don't have a chance to indicate errors in the ctor, so do it here. */
+    if (stp_dyninst_ctor_rc != 0) {
+	return stp_dyninst_ctor_rc;
+    }
+
+    return _stp_shm_connect(name);
 }
 
 int stp_dyninst_session_init(void)
 {
+    int rc;
+
     /* We don't have a chance to indicate errors in the ctor, so do it here. */
-    if (_stp_mem_fd < 0) {
-	return -errno;
+    if (stp_dyninst_ctor_rc != 0) {
+	return stp_dyninst_ctor_rc;
     }
 
-    int rc = systemtap_module_init();
+    /* Just in case stapdyn didn't do it (e.g. an old version), make sure our
+     * shared memory is initialized before we do anything else.  */
+    if (stp_dyninst_shm_init() == NULL)
+	return -ENOMEM;
+
+    rc = systemtap_module_init();
     if (rc == 0) {
 	stp_dyninst_master = getpid();
+	_stp_shm_finalize();
     }
     return rc;
 }
@@ -208,6 +305,9 @@ static void stp_dyninst_dtor(void)
 {
     stp_dyninst_session_exit();
 
+    _stp_print_cleanup();
+    _stp_shm_destroy();
+
     if (_stp_mem_fd != -1) {
 	close (_stp_mem_fd);
     }
