@@ -135,22 +135,38 @@ static struct
 
 // PR13489, inode-uprobes sometimes lacks the necessary SYMBOL_EXPORT's.
 #if !defined(STAPCONF_TASK_USER_REGSET_VIEW_EXPORTED)
-void *kallsyms_task_user_regset_view;
+static void *kallsyms_task_user_regset_view;
 #endif
 #if !defined(STAPCONF_UPROBE_REGISTER_EXPORTED)
-void *kallsyms_uprobe_register;
+static void *kallsyms_uprobe_register;
 #endif
 #if !defined(STAPCONF_UPROBE_UNREGISTER_EXPORTED)
-void *kallsyms_uprobe_unregister;
+static void *kallsyms_uprobe_unregister;
+#endif
+#if !defined(STAPCONF_URETPROBE_REGISTER_EXPORTED)
+static void *kallsyms_uretprobe_register;
+#endif
+#if !defined(STAPCONF_URETPROBE_UNREGISTER_EXPORTED)
+static void *kallsyms_uretprobe_unregister;
 #endif
 #if !defined(STAPCONF_UPROBE_GET_SWBP_ADDR_EXPORTED)
-void *kallsyms_uprobe_get_swbp_addr;
+static void *kallsyms_uprobe_get_swbp_addr;
 #endif
 
 /* task_work functions lack the necessary SYMBOL_EXPORT's */
 #if !defined(STAPCONF_TASK_WORK_ADD_EXPORTED)
-void *kallsyms_task_work_add;
-void *kallsyms_task_work_cancel;
+static void *kallsyms_task_work_add;
+static void *kallsyms_task_work_cancel;
+#endif
+
+#if !defined(STAPCONF_SIGNAL_WAKE_UP_STATE_EXPORTED)
+static void *kallsyms_signal_wake_up_state;
+#endif
+#if !defined(STAPCONF_SIGNAL_WAKE_UP_EXPORTED)
+static void *kallsyms_signal_wake_up;
+#endif
+#if !defined(STAPCONF___LOCK_TASK_SIGHAND_EXPORTED)
+static void *kallsyms___lock_task_sighand;
 #endif
 
 #include "access_process_vm.h"
@@ -161,12 +177,15 @@ void *kallsyms_task_work_cancel;
 #include "stp_string.c"
 #include "arith.c"
 #include "copy.c"
-#include "regs.c"
+#include "../regs.c"
 #include "regs-ia64.c"
 
 #if (defined(CONFIG_UTRACE) || defined(STAPCONF_UTRACE_VIA_TRACEPOINTS))
 #define HAVE_TASK_FINDER
 #include "task_finder.c"
+#else
+/* stub functionality that fails gracefully */
+#include "task_finder_stubs.c"
 #endif
 
 #include "sym.c"
@@ -187,9 +206,6 @@ void *kallsyms_task_work_cancel;
 struct unwind_context { };
 #endif
 
-static inline void _stp_runtime_entryfn_prologue(void) { }
-static inline void _stp_runtime_entryfn_epilogue(void) { }
-
 #ifdef module_param_cb			/* kernels >= 2.6.36 */
 #define _STP_KERNEL_PARAM_ARG const struct kernel_param
 #else
