@@ -725,7 +725,7 @@ adjustStartLoc (unsigned long startLoc,
   /* XXX - some, or all, of this should really be done by
      _stp_module_relocate and/or read_pointer. */
   dbug_unwind(2, "adjustStartLoc=%lx, ptrType=%s, m=%s, s=%s eh=%d\n",
-	      startLoc, _stp_eh_enc_name(ptrType), m->name, s->name, is_ehframe);
+	      startLoc, _stp_eh_enc_name(ptrType), m->path, s->name, is_ehframe);
   if (startLoc == 0
       || strcmp (m->name, "kernel")  == 0
       || (strcmp (s->name, ".absolute") == 0 && !is_ehframe))
@@ -1165,11 +1165,11 @@ static int unwind_frame(struct unwind_context *context,
 	if (unlikely(table_len == 0)) {
 		// Don't _stp_warn about this, debug_frame and/or eh_frame
 		// might actually not be there.
-		dbug_unwind(1, "Module %s: no unwind frame data", m->name);
+		dbug_unwind(1, "Module %s: no unwind frame data", m->path);
 		goto err;
 	}
 	if (unlikely(table_len & (sizeof(*fde) - 1))) {
-		_stp_warn("Module %s: frame_len=%d", m->name, table_len);
+		_stp_warn("Module %s: frame_len=%d", m->path, table_len);
 		goto err;
 	}
 
@@ -1181,12 +1181,12 @@ static int unwind_frame(struct unwind_context *context,
 		set_no_state_rule(i, Nowhere, state);
 
 	fde = _stp_search_unwind_hdr(pc, m, s, is_ehframe, user);
-	dbug_unwind(1, "%s: fde=%lx\n", m->name, (unsigned long) fde);
+	dbug_unwind(1, "%s: fde=%lx\n", m->path, (unsigned long) fde);
 
 	/* found the fde, now set startLoc and endLoc */
 	if (fde != NULL && is_fde(fde, table, table_len, is_ehframe)) {
 		cie = cie_for_fde(fde, table, table_len, is_ehframe);
-		dbug_unwind(1, "%s: cie=%lx\n", m->name, (unsigned long) cie);
+		dbug_unwind(1, "%s: cie=%lx\n", m->path, (unsigned long) cie);
 		if (likely(cie != NULL)) {
 			if (parse_fde_cie(fde, cie, table, table_len,
 					  &ptrType, user,
@@ -1215,7 +1215,7 @@ static int unwind_frame(struct unwind_context *context,
 	       if it didn't exist. These should never be missing except
 	       when there are toolchain bugs. */
 	    unsigned long tableSize;
-	    _stp_warn("No binary search table for %s frame, doing slow linear search for %s\n", (is_ehframe ? "eh" : "debug"), m->name);
+	    _stp_warn("No binary search table for %s frame, doing slow linear search for %s\n", (is_ehframe ? "eh" : "debug"), m->path);
 	    for (fde = table, tableSize = table_len; cie = NULL, tableSize > sizeof(*fde)
 		 && tableSize - sizeof(*fde) >= *fde; tableSize -= sizeof(*fde) + *fde, fde += 1 + *fde / sizeof(*fde)) {
 			dbug_unwind(3, "fde=%lx tableSize=%d\n", (long)*fde, (int)tableSize);
