@@ -434,6 +434,12 @@ void entry_op::print (ostream& o) const
 }
 
 
+void perf_op::print (ostream& o) const
+{
+  o << "@perf(" << *operand << ")";
+}
+
+
 void vardecl::print (ostream& o) const
 {
   o << name;
@@ -931,6 +937,10 @@ void stat_op::print (ostream& o) const
     case sc_max:
       o << "max(";
       break;
+
+    case sc_none:
+      assert (0); // should not happen, as sc_none is only used in foreach sorts
+      break;
     }
   stat->print(o);
   o << ")";
@@ -1037,7 +1047,21 @@ void foreach_loop::print (ostream& o) const
   o << "] in ";
   base->print_indexable (o);
   if (sort_direction != 0 && sort_column == 0)
-    o << (sort_direction > 0 ? "+" : "-");
+    {
+      switch (sort_aggr)
+        {
+        case sc_count: o << " @count"; break;
+        case sc_average: o << " @avg"; break;
+        case sc_min: o << " @min"; break;
+        case sc_max: o << " @max"; break;
+        case sc_sum: o << " @sum"; break;
+        case sc_none:
+        default: 
+          ;
+        }
+        
+      o << (sort_direction > 0 ? "+" : "-");
+    }
   if (limit)
     {
       o << " limit ";
@@ -1376,6 +1400,12 @@ array_in::visit (visitor* u)
   u->visit_array_in (this);
 }
 
+void
+regex_query::visit (visitor* u)
+{
+  u->visit_regex_query (this);
+}
+
 void
 comparison::visit (visitor* u)
 {
@@ -1451,6 +1481,13 @@ entry_op::visit (visitor* u)
 }
 
 
+void
+perf_op::visit (visitor* u)
+{
+  u->visit_perf_op(this);
+}
+
+
 void
 arrayindex::visit (visitor* u)
 {
@@ -1798,6 +1835,13 @@ traversing_visitor::visit_array_in (array_in* e)
   e->operand->visit (this);
 }
 
+void
+traversing_visitor::visit_regex_query (regex_query* e)
+{
+  e->left->visit (this);
+  e->right->visit (this); // TODOXXX do we need to traverse the literal in RHS?
+}
+
 void
 traversing_visitor::visit_comparison (comparison* e)
 {
@@ -1858,6 +1902,13 @@ traversing_visitor::visit_entry_op (entry_op* e)
 }
 
 
+void
+traversing_visitor::visit_perf_op (perf_op* e)
+{
+  e->operand->visit (this);
+}
+
+
 void
 traversing_visitor::visit_arrayindex (arrayindex* e)
 {
@@ -2056,6 +2107,13 @@ varuse_collecting_visitor::visit_entry_op (entry_op *e)
 }
 
 
+void
+varuse_collecting_visitor::visit_perf_op (perf_op *e)
+{
+  functioncall_traversing_visitor::visit_perf_op (e);
+}
+
+
 void
 varuse_collecting_visitor::visit_print_format (print_format* e)
 {
@@ -2440,6 +2498,12 @@ throwing_visitor::visit_array_in (array_in* e)
   throwone (e->tok);
 }
 
+void
+throwing_visitor::visit_regex_query (regex_query* e)
+{
+  throwone (e->tok);
+}
+
 void
 throwing_visitor::visit_comparison (comparison* e)
 {
@@ -2495,6 +2559,13 @@ throwing_visitor::visit_entry_op (entry_op* e)
 }
 
 
+void
+throwing_visitor::visit_perf_op (perf_op* e)
+{
+  throwone (e->tok);
+}
+
+
 void
 throwing_visitor::visit_arrayindex (arrayindex* e)
 {
@@ -2699,6 +2770,14 @@ update_visitor::visit_array_in (array_in* e)
   provide (e);
 }
 
+void
+update_visitor::visit_regex_query (regex_query* e)
+{
+  replace (e->left);
+  replace (e->right); // TODOXXX do we need to replace literal in RHS?
+  provide (e);
+}
+
 void
 update_visitor::visit_comparison (comparison* e)
 {
@@ -2767,6 +2846,13 @@ update_visitor::visit_entry_op (entry_op* e)
   provide (e);
 }
 
+void
+update_visitor::visit_perf_op (perf_op* e)
+{
+  replace (e->operand);
+  provide (e);
+}
+
 void
 update_visitor::visit_arrayindex (arrayindex* e)
 {
@@ -2970,6 +3056,12 @@ deep_copy_visitor::visit_array_in (array_in* e)
   update_visitor::visit_array_in(new array_in(*e));
 }
 
+void
+deep_copy_visitor::visit_regex_query (regex_query* e)
+{
+  update_visitor::visit_regex_query(new regex_query(*e));
+}
+
 void
 deep_copy_visitor::visit_comparison (comparison* e)
 {
@@ -3028,6 +3120,12 @@ deep_copy_visitor::visit_entry_op (entry_op* e)
   update_visitor::visit_entry_op(new entry_op(*e));
 }
 
+void
+deep_copy_visitor::visit_perf_op (perf_op* e)
+{
+  update_visitor::visit_perf_op(new perf_op(*e));
+}
+
 void
 deep_copy_visitor::visit_arrayindex (arrayindex* e)
 {
