@@ -26,6 +26,7 @@
 #include <ext/stdio_filebuf.h>
 
 extern "C" {
+#include <elf.h>
 #include <fcntl.h>
 #include <grp.h>
 #include <pwd.h>
@@ -1007,6 +1008,25 @@ normalize_machine(const string& machine)
   return machine;
 }
 
+int
+elf_class_from_normalized_machine (const string &machine)
+{
+  // Must match kernel machine architectures as used un tapset directory.
+  // And must match normalization done in normalize_machine ().
+  if (machine == "i386"
+      || machine == "arm")         // arm assumes 32-bit
+    return ELFCLASS32;
+  else if (machine == "s390"       // powerpc and s390 always assume 64-bit,
+           || machine == "powerpc" // see normalize_machine ().
+           || machine == "x86_64"
+           || machine == "ia64")
+    return ELFCLASS64;
+
+  cerr << _F("Unknown kernel machine architecture '%s', don't know elf class",
+             machine.c_str()) << endl;
+  return -1;
+}
+
 string
 kernel_release_from_build_tree (const string &kernel_build_tree, int verbose)
 {
@@ -1063,4 +1083,34 @@ get_self_path()
   return std::string(file);
 }
 
+
+#ifndef HAVE_PPOLL
+// This is a poor-man's ppoll, only used carefully by readers that need to be
+// interruptible, like remote::run and mutator::run.  It does not provide the
+// same guarantee of atomicity as on systems with a true ppoll.
+//
+// In our use, this would cause trouble if a signal came in any time from the
+// moment we mask signals to prepare pollfds, to the moment we call poll in
+// emulation here.  If there's no data on any of the pollfds, we will be stuck
+// waiting indefinitely.
+//
+// Since this is mainly about responsiveness of CTRL-C cleanup, we'll just
+// throw in a one-second forced timeout to ensure we have a chance to notice
+// there was an interrupt without too much delay.
+int
+ppoll(struct pollfd *fds, nfds_t nfds,
+      const struct timespec *timeout_ts,
+      const sigset_t *sigmask)
+{
+  sigset_t origmask;
+  int timeout = (timeout_ts == NULL) ? 1000 // don't block forever...
+    : (timeout_ts->tv_sec * 1000 + timeout_ts->tv_nsec / 1000000);
+  sigprocmask(SIG_SETMASK, sigmask, &origmask);
+  int rc = poll(fds, nfds, timeout);
+  sigprocmask(SIG_SETMASK, &origmask, NULL);
+  return rc;
+}
+#endif
+
+
 /* vim: set sw=2 ts=8 cino=>4,n-2,{2,^-2,t0,(0,u0,w1,M1 : */
