@@ -2,6 +2,7 @@
 #define UTIL_H
 
 #include "config.h"
+#include <cstdlib>
 #include <cstring>
 #include <cerrno>
 #include <string>
@@ -20,6 +21,7 @@ extern "C" {
 #include <stdint.h>
 #include <spawn.h>
 #include <assert.h>
+#include <poll.h>
 }
 
 #include "privilege.h"
@@ -81,6 +83,7 @@ std::string escape_glob_chars (const std::string& str);
 std::string unescape_glob_chars (const std::string& str);
 std::string kernel_release_from_build_tree (const std::string &kernel_build_tree, int verbose = 0);
 std::string normalize_machine(const std::string& machine);
+int elf_class_from_normalized_machine(const std::string& machine);
 std::string autosprintf(const char* format, ...) __attribute__ ((format (printf, 1, 2)));
 const std::set<std::string>& localization_variables();
 std::string get_self_path();
@@ -269,6 +272,29 @@ struct stap_sigmasker {
       }
 };
 
+// Convert a possibly-relative path to a full path
+inline std::string
+resolve_path(const std::string& path)
+{
+  std::string result(path);
+  char* resolved_path = realpath(path.c_str(), NULL);
+  if (resolved_path)
+    {
+      result = resolved_path;
+      std::free(resolved_path);
+    }
+  return result;
+}
+
+
+#ifndef HAVE_PPOLL
+// This is a poor-man's ppoll; see the implementation for more details...
+int ppoll(struct pollfd *fds, nfds_t nfds,
+          const struct timespec *timeout_ts,
+          const sigset_t *sigmask);
+#endif
+
+
 #endif // UTIL_H
 
 /* vim: set sw=2 ts=8 cino=>4,n-2,{2,^-2,t0,(0,u0,w1,M1 : */
